#! /usr/bin/python

# code written by KTL
# * skydome and referenced textures coming from Optic Delusion
# * referenced hills texture from Blaster Wisconsin
# * use this code freely to make life more complicated.
# * everyone free to deliver me pizza, no beer.

# code
# * usage: python convert.py <inputmap.bzw> [<outputmap.bzw>]
#   if no outputmap specified it writes to out.bzw
# * the code runs through old style bzw files
# * boxes and pyramids become meshes
# * texture coordinates are put in such a way that textures from overlapping faces fit exactly on each other.
#   this is true for horizontal faces (the easy ones) and for box-sides
# * elevated teleporters get a small doorstep box
#   some buggy code does not allow passing through porters on elevated meshes.
# * no normals are inserted since boxes and pyramids seem to have trivial normals.
# * skydome and hills are placed around the map.
# * inside points are placed, 8 per box close to the corners, 5 per pyramid.
# * places 4 banners of the local dragon/god/pet, they fit nice on the modified mad city map.
# * 2 types of pyramids: the ones that are elevated or drivethrough are different.
# * it adds some default options
# * if it should copy what it does not touch (not box,pyr,tele,link) then uncomment the last 2 lines.

# bugs
# ... ?

# todo
# * better handling of secondary attributes like shootthrough and drivethrough
# * add more decorations (shootthrough/drivethrough)
# * change the sky periodically with dynamic colors, there are nice textures available.
# * there are pyramids with a negative height too...
# * ...

from math import *
import sys
from bzwfragments import *

filename = 'mmcbgow.bzw'
outfilename = 'out.bzw'
#filename = '/home/donf/src/bzflagspiralworld/example.bzw'

for arg in sys.argv:
    print arg

if len(sys.argv) > 1:
    filename = sys.argv[1]
if len(sys.argv) > 2:
    outfilename = sys.argv[2]

out=open(outfilename, 'w')


#comment to avoid adding more options
print >>out, options
#comment to avoid adding some default materials
print >>out, materials
#comment to get rid of the pet
print >>out, hashmarkdragon
#comment to get rid of the skydome but dont forget to replace it
print >>out, skydome
#comment to get rid of the surrounding hills.
print >>out, mountains



class Data:
    position = [0,0,0]
    size = [1,1,1]
    rotation = 0
    drivethrough = 0
    shootthrough = 0
    border = 1.0
    flipz = 0

    def add(self,split):
        if split[0] == 'position':
            self.position = [float(split[1]),float(split[2]),float(split[3])]
        elif split[0] == 'rotation':
            self.rotation = float(split[1])
        elif split[0] == 'size':
            self.size = [float(split[1]),float(split[2]),float(split[3])]
        elif split[0] == 'border':
            self.border= float(split[1])
        elif split[0] == 'drivethrough':
            self.drivethrough = 1
        elif split[0] == 'shootthrough':
            self.shootthrough = 1
        elif split[0] == 'passable':
            self.shootthrough = 1
            self.drivethrough = 1
        elif split[0] == 'flipz':
            self.flipz = 1


def getDistanceFromSomeAxis(c1x,c1y,c2x,c2y):

    c1x = float(c1x)
    c1y = float(c1y)
    c2x = float(c2x)
    c2y = float(c2y)

    if abs(c1x - c2x) < 0.000001:
        return c1y
    elif abs(c1y - c2y) < 0.0000001:
        return c1x
    else:
        d1y = 0.0

        c12bnum    = c1x * (c2y - c1y)
        c12bdenum  = (c2x - c1x)

        if c12bdenum == 0:

            d1y = c1y  # getDistanceFromSomeAxis(c1y,c1x,c2y,c2x)

        else:
            yint = c1y - (c12bnum / c12bdenum)
            d1y = sqrt((c1y - yint) * (c1y - yint) + c1x * c1x)
            if c1x < 0:
                d1y = - d1y
        return d1y


def printbox(data,out):
    upz = data.position[2] + data.size[2]
    downz = data.position[2] #- data.size[2]
    data.rotation = data.rotation % 180
    rot = data.rotation * pi / 180
    rotmat = [[cos(rot),-sin(rot)],[sin(rot),cos(rot)]]

    c1x = rotmat[0][0] * data.size[0] + rotmat[0][1] * data.size[1] + data.position[0]
    c1y = rotmat[1][0] * data.size[0] + rotmat[1][1] * data.size[1] + data.position[1]

    c2x = rotmat[0][0] * data.size[0] + rotmat[0][1] * -data.size[1] + data.position[0]
    c2y = rotmat[1][0] * data.size[0] + rotmat[1][1] * -data.size[1] + data.position[1]

    c3x = rotmat[0][0] * -data.size[0] + rotmat[0][1] * -data.size[1] + data.position[0]
    c3y = rotmat[1][0] * -data.size[0] + rotmat[1][1] * -data.size[1] + data.position[1]

    c4x = rotmat[0][0] * -data.size[0] + rotmat[0][1] * data.size[1] + data.position[0]
    c4y = rotmat[1][0] * -data.size[0] + rotmat[1][1] * data.size[1] + data.position[1]

    n1xn = rotmat[0][0] * data.size[0]
    n1yn = rotmat[1][0] * data.size[0]
    deler = sqrt(n1xn * n1xn + n1yn * n1yn)
    n1xn = n1xn / deler
    n1yn = n1yn / deler

    n2xn = rotmat[0][1] * data.size[1]
    n2yn = rotmat[1][1] * data.size[1]
    deler = sqrt(n2xn * n2xn + n2yn * n2yn)
    n2xn = n2xn / deler
    n2yn = n2yn / deler


 #   print >>out, "box"
 #   print >>out, "position %f %f %f" % (data.position[0],data.position[1],upz)
 #   print >>out, "size %f %f %f" % (data.size[0],data.size[1],0.01)
 #   print >>out, "rotation %f " %  data.rotation
#    print >>out, "top mtransp"
 #   print >>out, "end"


    print >>out, "mesh\n\tname bzbox"

    if data.drivethrough == 0:
        ind = 0.01 # this influences the distance away from the corners for the inside points.
        print >>out, "\tinside %f %f %f" % (c4x - ind * (c4x - c2x) ,c4y - ind * (c4y - c2y), upz   - (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c4x - ind * (c4x - c2x) ,c4y - ind * (c4y - c2y), downz + (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c2x - ind * (c2x - c4x) ,c2y - ind * (c2y - c4y), upz   - (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c2x - ind * (c2x - c4x) ,c2y - ind * (c2y - c4y), downz + (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c3x - ind * (c3x - c1x) ,c3y - ind * (c3y - c1y), upz   - (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c3x - ind * (c3x - c1x) ,c3y - ind * (c3y - c1y), downz + (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c1x - ind * (c1x - c3x) ,c1y - ind * (c1y - c3y), upz   - (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c1x - ind * (c1x - c3x) ,c1y - ind * (c1y - c3y), downz + (ind * data.size[2]))
        print >>out, "\t"

    print >>out, "\tvertex %f %f %f" % (c4x,c4y,upz)
    print >>out, "\tvertex %f %f %f" % (c3x,c3y,upz)

    print >>out, "\tvertex %f %f %f" % (c2x,c2y,upz)
    print >>out, "\tvertex %f %f %f" % (c1x,c1y,upz)
    print >>out, "\tvertex %f %f %f" % (c1x,c1y,downz)
    print >>out, "\tvertex %f %f %f" % (c2x,c2y,downz)

    print >>out, "\tvertex %f %f %f" % (c3x,c3y,downz)
    print >>out, "\tvertex %f %f %f" % (c4x,c4y,downz)
    print >>out, "\t"

    rx = data.size[0] / 4
    ry = data.size[1] / 4
    rz = data.size[2] / 8

    #for the top and bottom of the boxes
    print >>out, "\ttexcoord %f %f" % ((c4x ) / 4.0,(c4y ) / 4.0)
    print >>out, "\ttexcoord %f %f" % ((c3x ) / 4.0,(c3y ) / 4.0)
    print >>out, "\ttexcoord %f %f" % ((c2x ) / 4.0,(c2y ) / 4.0)
    print >>out, "\ttexcoord %f %f" % ((c1x ) / 4.0,(c1y ) / 4.0)

    #to make the side textures overlap just as the top textures in order to avoid problems with the graphics card...
    #i got to extend the sides till they hit the axes, and start texturing from one of them.
    #in other words: given 2 points shoot a line through them and calculate the distance till it crosses the axes.

    dc12 = getDistanceFromSomeAxis(c1x,c1y,c2x,c2y) / 4.0
    dc21 = getDistanceFromSomeAxis(c2x,c2y,c1x,c1y) / 4.0
    dc23 = getDistanceFromSomeAxis(c2x,c2y,c3x,c3y) / 4.0
    dc32 = getDistanceFromSomeAxis(c3x,c3y,c2x,c2y) / 4.0
    dc34 = getDistanceFromSomeAxis(c3x,c3y,c4x,c4y) / 4.0
    dc43 = getDistanceFromSomeAxis(c4x,c4y,c3x,c3y) / 4.0
    dc41 = getDistanceFromSomeAxis(c4x,c4y,c1x,c1y) / 4.0
    dc14 = getDistanceFromSomeAxis(c1x,c1y,c4x,c4y) / 4.0



    #side from c1 to c2

    #    print [dc12, dc21,dc23,dc32,dc34,dc43,dc41,dc14]

    dzn = downz / 4.0
    uzn = upz   / 4.0
    #texcoord nr 4-7
    print >>out, "\ttexcoord %f %f" % (dc12,uzn)
    print >>out, "\ttexcoord %f %f" % (dc21,uzn)
    print >>out, "\ttexcoord %f %f" % (dc21,dzn)
    print >>out, "\ttexcoord %f %f" % (dc12,dzn)
    #texcoord nr 8-11
    print >>out, "\ttexcoord %f %f" % (dc23,uzn)
    print >>out, "\ttexcoord %f %f" % (dc32,uzn)
    print >>out, "\ttexcoord %f %f" % (dc32,dzn)
    print >>out, "\ttexcoord %f %f" % (dc23,dzn)
    #texcoord nr 12-15
    print >>out, "\ttexcoord %f %f" % (dc34,uzn)
    print >>out, "\ttexcoord %f %f" % (dc43,uzn)
    print >>out, "\ttexcoord %f %f" % (dc43,dzn)
    print >>out, "\ttexcoord %f %f" % (dc34,dzn)
    #texcoord nr 16-19
    print >>out, "\ttexcoord %f %f" % (dc14,uzn)
    print >>out, "\ttexcoord %f %f" % (dc41,uzn)
    print >>out, "\ttexcoord %f %f" % (dc41,dzn)
    print >>out, "\ttexcoord %f %f" % (dc14,dzn)

    if data.drivethrough == 1 and data.shootthrough == 1:
        print >>out, boxmesh_sd_end
    elif data.drivethrough == 1:
        print >>out, boxmesh_d_end
    elif data.shootthrough == 1:
        print >>out, boxmesh_s_end
    else:
        print >>out, boxmesh_end


def printpyramid(data,out):

    if data.flipz  == 0 :
        upz = data.position[2] + data.size[2]
        downz = data.position[2]
    else:
        upz = data.position[2]
        downz = data.position[2] + data.size[2]

#    if position[2] > 0:
#       data.drivethrough = 1

    rot = data.rotation * pi / 180
    rotmat = [[cos(rot),-sin(rot)],[sin(rot),cos(rot)]]

    c1x = rotmat[0][0] * data.size[0] + rotmat[0][1] * data.size[1] + data.position[0]
    c1y = rotmat[1][0] * data.size[0] + rotmat[1][1] * data.size[1] + data.position[1]

    c2x = rotmat[0][0] * data.size[0] + rotmat[0][1] * -data.size[1] + data.position[0]
    c2y = rotmat[1][0] * data.size[0] + rotmat[1][1] * -data.size[1] + data.position[1]

    c3x = rotmat[0][0] * -data.size[0] + rotmat[0][1] * -data.size[1] + data.position[0]
    c3y = rotmat[1][0] * -data.size[0] + rotmat[1][1] * -data.size[1] + data.position[1]

    c4x = rotmat[0][0] * -data.size[0] + rotmat[0][1] * data.size[1] + data.position[0]
    c4y = rotmat[1][0] * -data.size[0] + rotmat[1][1] * data.size[1] + data.position[1]

    n1xn = rotmat[0][0] * data.size[0]
    n1yn = rotmat[1][0] * data.size[0]
    deler = sqrt(n1xn * n1xn + n1yn * n1yn)
    n1xn = n1xn / deler
    n1yn = n1yn / deler

    n2xn = rotmat[0][1] * data.size[1]
    n2yn = rotmat[1][1] * data.size[1]
    deler = sqrt(n2xn * n2xn + n2yn * n2yn)
    n2xn = n2xn / deler
    n2yn = n2yn / deler

    print >>out, "mesh\n\tname bzpyr"
    if data.drivethrough == 0 and data.position[2] == 0:
        ind = 0.01
        print >>out, "\tinside %f %f %f" % (data.position[0] ,data.position[1], upz - (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c4x - ind * (c4x - c2x) ,c4y - ind * (c4y - c2y), downz + (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c2x - ind * (c2x - c4x) ,c2y - ind * (c2y - c4y), downz + (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c3x - ind * (c3x - c1x) ,c3y - ind * (c3y - c1y), downz + (ind * data.size[2]))
        print >>out, "\tinside %f %f %f" % (c1x - ind * (c1x - c3x) ,c1y - ind * (c1y - c3y), downz + (ind * data.size[2]))
        print >>out, ""


    if data.flipz  == 0 :
        print >>out, "\tvertex %f %f %f" % (c1x,c1y,downz)
        print >>out, "\tvertex %f %f %f" % (c2x,c2y,downz)
        print >>out, "\tvertex %f %f %f" % (c3x,c3y,downz)
        print >>out, "\tvertex %f %f %f" % (c4x,c4y,downz)
        print >>out, "\tvertex %f %f %f" % (data.position[0],data.position[1],upz)
    else:
        print >>out, "\tvertex %f %f %f" % (c4x,c4y,downz)
        print >>out, "\tvertex %f %f %f" % (c3x,c3y,downz)
        print >>out, "\tvertex %f %f %f" % (c2x,c2y,downz)
        print >>out, "\tvertex %f %f %f" % (c1x,c1y,downz)
        print >>out, "\tvertex %f %f %f" % (data.position[0],data.position[1],upz)

    print >>out, ""
    #    print >>out, "\tnormal %f %f %f" % (n1xn,n1yn,0)
    #    print >>out, "\tnormal %f %f %f" % (n2xn,n2yn,0)
    #    print >>out, "\tnormal %f %f %f" % (-n1xn,-n1yn,0)
    #    print >>out, "\tnormal %f %f %f" % (-n2xn,-n2yn,0)
    #    print >>out, "\tnormal %f %f %f" % (0,0,-1)



    dc12 = getDistanceFromSomeAxis(c1x,c1y,c2x,c2y) / 4.0
    dc21 = getDistanceFromSomeAxis(c2x,c2y,c1x,c1y) / 4.0
    dc23 = getDistanceFromSomeAxis(c2x,c2y,c3x,c3y) / 4.0
    dc32 = getDistanceFromSomeAxis(c3x,c3y,c2x,c2y) / 4.0
    dc34 = getDistanceFromSomeAxis(c3x,c3y,c4x,c4y) / 4.0
    dc43 = getDistanceFromSomeAxis(c4x,c4y,c3x,c3y) / 4.0
    dc41 = getDistanceFromSomeAxis(c4x,c4y,c1x,c1y) / 4.0
    dc14 = getDistanceFromSomeAxis(c1x,c1y,c4x,c4y) / 4.0


    rx = data.size[0] / 4
    ry = data.size[1] / 4
    rz = data.size[2] / 8

    #top of the pyr

    #3 different projections possible. horizontal (default), vertical (flat pyramids), or straight on the surface, taking into account the length of the steep surface

    showpyramidwallsashorizontalprojection = 1
    if showpyramidwallsashorizontalprojection == 1:
        #        print >>out, "\ttexcoord %f %f" % (0,rz)
        #        print >>out, "\ttexcoord %f %f" % (0,rz)
        print >>out, "\ttexcoord %f %f" % ( (dc12 + dc21) / 4.0 ,rz)
        print >>out, "\ttexcoord %f %f" % ( (dc23 + dc32) / 4.0 ,rz)        
        print >>out, "\ttexcoord %f %f" % ( (dc34 + dc43) / 4.0 ,rz)
        print >>out, "\ttexcoord %f %f" % ( (dc41 + dc14) / 4.0 ,rz)
    else:        
        ly = sqrt(rx * rx + rz * rz)
        lx = sqrt(ry * ry + rz * rz)
        print >>out, "\ttexcoord %f %f" % (0,ly)
        print >>out, "\ttexcoord %f %f" % (0,lx)
        print >>out, "\ttexcoord %f %f" % (0,ly)
        print >>out, "\ttexcoord %f %f" % (0,lx)

    #    print >>out, "\ttexcoord %f %f" % (dc23 / 2.0, 0)
    #    print >>out, "\ttexcoord %f %f" % (dc32 / 2.0, 0)
    
    #    print >>out, "\ttexcoord %f %f" % (dc21 / 2.0, 0)
    #    print >>out, "\ttexcoord %f %f" % (dc12 / 2.0, 0)
    
    print >>out, "\ttexcoord %f %f" % (dc21 / 2.0,0)
    print >>out, "\ttexcoord %f %f" % (dc12 / 2.0,0)
    #texcoord nr 8-11
    print >>out, "\ttexcoord %f %f" % (dc32 / 2.0,0)
    print >>out, "\ttexcoord %f %f" % (dc23 / 2.0,0)
    #texcoord nr 12-15
    print >>out, "\ttexcoord %f %f" % (dc43 / 2.0,0)
    print >>out, "\ttexcoord %f %f" % (dc34 / 2.0,0)
    #texcoord nr 16-19
    print >>out, "\ttexcoord %f %f" % (dc14 / 2.0,0)
    print >>out, "\ttexcoord %f %f" % (dc41 / 2.0,0)




    #for the bottom of the pyrs, aligned with bottom faces from boxes
    print >>out, "\ttexcoord %f %f" % (c4x / 4.0,c4y / 4.0)
    print >>out, "\ttexcoord %f %f" % (c3x / 4.0,c3y / 4.0)
    print >>out, "\ttexcoord %f %f" % (c2x / 4.0,c2y / 4.0)
    print >>out, "\ttexcoord %f %f" % (c1x / 4.0,c1y / 4.0)

    if data.drivethrough == 1 and data.shootthrough == 1:
        print >>out, pyrmesh_sd_end
    elif data.drivethrough == 1:
        print >>out, pyrmesh_d_end
    elif data.shootthrough == 1:
        print >>out, pyrmesh_s_end
    else:
        print >>out, pyrmesh_end


def printteleporter(data,out):
    if data.position[2] == 0:
#        print >>out, 'teleporter'
        print >>out, "position %f %f %f" % (data.position[0],data.position[1],data.position[2])
        print >>out, "\tsize %f %f %f" % (data.size[0],data.size[1],data.size[2])
        print >>out, "\trotation %f " %  data.rotation
        print >>out, "\tborder %f " % data.border
        print >>out, 'end'
    else:
        #put a small box beneath elevated porters as a doorstep to circumvent some buggy code that would otherwise prevent a tank from passing
        elevation = 0.09
#        print >>out, 'teleporter'
        print >>out, "position %f %f %f" % (data.position[0],data.position[1],data.position[2] + elevation)
        print >>out, "\tsize %f %f %f" % (data.size[0],data.size[1],data.size[2])
        print >>out, "\trotation %f " % data.rotation
        print >>out, "\tborder %f " % data.border
        print >>out, 'end'

        print >>out, "box"
        print >>out, "\tposition %f %f %f" % (data.position[0],data.position[1],data.position[2])
        print >>out, "\tsize %f %f %f" % (data.size[0],data.size[1],elevation)
        print >>out, "\trotation %f " %  data.rotation
        print >>out, "end"


def parsefile(filename,out):

    f=open(filename, 'r')


    lines = f.readlines()
    counter = 0
    currentstate = "root"
    data = Data()
    for line in lines:
        split = line.split()

        if split != []:
            if line[0] == '#':
                print >>out, line.replace('\n','')
            elif split[0] == 'end':
                counter = counter + 1
                if currentstate == 'box':
                    print >>out, "#end"
                    printbox(data,out)
                elif currentstate == 'pyramid':
                    print >>out, "#end"
                    printpyramid(data,out)
                elif currentstate == 'teleporter':
                    print >>out, "#end"
                    printteleporter(data,out)
                elif currentstate == 'link':
                    print >>out, 'end'
                currentstate = 'root'
                data = Data()
            elif split[0] == 'box':
                currentstate = 'box'
                print >>out, "#%s" % line.replace('\n','')
            elif split[0] == 'pyramid':
                currentstate = 'pyramid'
                print >>out, "#%s" % line.replace('\n','')
            elif split[0] == 'teleporter':
                currentstate = 'teleporter'
                print >>out, line.replace('\n','')
            elif split[0] == 'link':
                currentstate = 'link'
                print >>out, 'link'
            else:
                if currentstate == 'box':
                    data.add(split)
                    print >>out, "#%s" % line.replace('\n','')
                elif currentstate == 'pyramid':
                    data.add(split)
                    print >>out, "#%s" % line.replace('\n','')
                elif currentstate == 'teleporter':
                    data.add(split)
                    print >>out, "#%s" % line.replace('\n','')
                elif currentstate == 'link':
                    print >>out, line.replace('\n','')
                else:
                    print >>out, "#%s" % line.replace('\n','')
#                    print >>out, line




parsefile(filename,out)


